#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#define LOCAL_SOCKET	1
#ifndef FD_ALLOC
#define FD_ALLOC(nfds) ((fd_set*) malloc((nfds+7) / 8 ) )
#endif

extern void debug(const char *fmt, ...);
extern void error(const char *fmt, ...);


//Return values:
// 1: Error (in select)
// 2: Server has disconnected
// 3: Viewer has disconnected
// 4: Error when reading from viewer 
// 5: Error when reading from server
int do_repeater(int server, int viewer)
{

        /** vars for viewer input data **/
    char viewerbuf[1024];       /* viewer input buffer */
    int viewerbuf_len;          /* available data in viewerbuf */

        /** vars for server input data **/
    char serverbuf[1024];       /* server input buffer */
    int serverbuf_len;          /* available data in serverbuf */

        /** other variables **/
    int nfds, len;
    fd_set *ifds, *ofds;
    
    /*
     * repeater between server and viewer 
     */
    nfds = ((viewer < server) ? server : viewer) + 1;
    ifds = FD_ALLOC(nfds);
    ofds = FD_ALLOC(nfds);
    viewerbuf_len = 0;
    serverbuf_len = 0;

    while (true) {
        FD_ZERO(ifds);
        FD_ZERO(ofds);

                /** prepare for reading viewer input **/
        if ((viewerbuf_len < (int) sizeof(viewerbuf))) {
            FD_SET(viewer, ifds);
        }

                /** prepare for reading server input **/
        if ((serverbuf_len < (int) sizeof(serverbuf))) {
            FD_SET(server, ifds);
        }

        if (select(nfds, ifds, ofds, NULL, NULL) == -1) {

            /*
             * error in select, parent process should close both connections  
             */
            error("do_repeater(): select() failed, errno=%d (%s)\n", errno, strerror(errno));
            return 1;
        }

        /*
         * server => viewer 
         */
        if (FD_ISSET(server, ifds)
            && (serverbuf_len < (int) sizeof(serverbuf))) {
            len = recv(server, serverbuf + serverbuf_len, sizeof(serverbuf) - serverbuf_len, 0);

            if (len == 0) {
                debug("do_repeater(): connection closed by server\n");
                
                //When server closes connection, parent process should also disconnect viewer
                return 2;
            }
            else if (len == -1) {
                if (errno != ECONNRESET) {

                    /*
                     * error 
                     */
                    debug("do_repeater(): recv() from server failed, errno=%d (%s)\n", errno, strerror(errno));
                    return 5;
                }
                else {
                    debug("do_repeater(): ECONNRESET detected\n");
                }
            }
            else {
                serverbuf_len += len;
            }
        }

        /*
         * viewer => server 
         */
        if (FD_ISSET(viewer, ifds)
            && (viewerbuf_len < (int) sizeof(viewerbuf))) {
            len = recv(viewer, viewerbuf + viewerbuf_len, sizeof(viewerbuf) - viewerbuf_len, 0);

            if (len == 0) {
                debug("do_repeater(): connection closed by viewer\n");

                //When viewer closes connection,parent process should also disconnect server 
                return 3;
            }
            else if (len == -1) {

                /*
                 * error on reading from viewer 
                 */
                debug("do_repeater(): recv() from viewer failed, errno = %d (%s)\n", errno, strerror(errno));
                return 4;
            }
            else {

                /*
                 * repeat 
                 */
                viewerbuf_len += len;
            }
        }

        /*
         * flush data in viewerbuffer to server 
         */
        if (0 < viewerbuf_len) {
            len = send(server, viewerbuf, viewerbuf_len, 0);

            if (len == -1) {
                debug("do_repeater(): send() (to server) failed, errno=%d (%s)\n", errno, strerror(errno));
            }
            else if (0 < len) {

                /*
                 * move data on to top of buffer 
                 */
                viewerbuf_len -= len;

                if (0 < viewerbuf_len)
                    memcpy(viewerbuf, viewerbuf + len, viewerbuf_len);

                assert(0 <= viewerbuf_len);
            }
        }

        /*
         * flush data in serverbuffer to viewer 
         */
        if (0 < serverbuf_len) {
            len = send(viewer, serverbuf, serverbuf_len, 0);

            if (len == -1) {
                debug("do_repeater(): send() (to viewer) failed, errno=%d (%s)\n", errno, strerror(errno));
            }
            else {
                serverbuf_len -= len;

                if (len < serverbuf_len)
                    memcpy(serverbuf, serverbuf + len, serverbuf_len);

                assert(0 <= serverbuf_len);
            }
        }
    }
}
